module Graph where

import Data.List

{-
    Următoarele exerciții vor avea drept scop implementarea unei mici biblioteci
    pentru grafuri ORIENTATE.
    După cum știți, există mai multe modalități de reprezentare a unui graf.
    Biblioteca noastră va defini mai multe astfel de reprezentări, precum
    și algoritmi care operează pe grafuri.
    Algoritmii vor fi independenți de reprezentarea internă a grafului - ei
    vor funcționa indiferent de ce structură de date am ales noi pentru un anume
    graf.
    Pentru a obține această genericitate vom abstractiza noțiunea de graf într-o
    clasă care va expune operațiile pe care orice structură de date de tip graf
    ar trebui să le aibă.
-}

-- reprezentăm nodurile ca întregi
type Node = Int

-- reprezentăm arcele ca perechi de noduri
type Arc = (Node, Node)


-------------------------------------------------------------------------------

{-
    1.(1.5p) Analizați clasa Graph definită mai jos și scrieți implementările
    implicite pentru funcțiile din această clasă.

    Clasa Graph definește interfața pentru toate structurile de grafuri
    pe care le vom implementa mai jos.
-}
class Graph g where

    -- Construiește un graf plecând de la o listă de noduri și arcele dintre noduri
    build :: [Node] -> [Arc] -> g

    -- Lista tuturor nodurilor din graf
    nodes :: g -> [Node] -- lista nodurilor din graf

    -- Lista arcelor din graf
    arcs :: g -> [Arc] -- lista muchiilor din graf

    -- Lista nodurilor către care nodul dat ca parametru are un arc
    nodeOut :: g -> Node -> [Node]

    -- Lista nodurilor care au un arc către nodul dat ca parametru
    nodeIn :: g -> Node -> [Node]

    -- Verifică dacă există un arc între două noduri
    arcExists :: g -> Node -> Node -> Bool

    -- TODO: implementari default
    arcExists g a b = undefined
    arcs g = undefined
    nodeIn g n = undefined
    nodeOut g n = undefined

-------------------------------------------------------------------------------

{-
    2.(2p) Definiți tipul AdjListGraph care reprezintă un graf ca pe o serie de
    perechi (nod, listă vecini). Includeți AdjListGraph în clasa Graph
-}
-- TODO: de rafinat parametrul lui ALGraph
newtype AdjListGraph = ALGraph Node deriving Show

instance Graph AdjListGraph where
  -- TODO

-------------------------------------------------------------------------------

{-
    3. (1p) Definiți tipul ArcGraph care reprezintă un graf ca o listă de noduri
    și o listă de arce între noduri. Includeți ArcGraph în clasa Graph
-}
-- TODO: de rafinat constructorul de date AGraph
data ArcGraph = AGraph deriving Show

instance Graph ArcGraph where
  -- TODO

-------------------------------------------------------------------------------

{-
    4. (0.5p) Definiți funcția convert care face conversia între
    reprezentări diferite de grafuri.
-}
convert :: (Graph g1, Graph g2) => g1 -> g2
convert g1 = undefined -- TODO


-------------------------------------------------------------------------------
{-
    O traversare a unui graf este o listă de perechi (nod, Maybe
    părinteNode). Această structură va conține toate nodurile rezultate în
    urma unei parcurgeri a unui graf, în ordinea apariției în parcurgere și
    împreuna cu părintele nodului din parcurgere (Pentru un nod N, părintele
    său este nodul din care s-a ajuns la N în decursul parcurgerii)
-}
type Traversal = [(Node, Maybe Node)]

-- O cale în graf este reprezentată ca listă de noduri din acea cale.
type Path = [Node]

{-
    Definiție pentru algoritmi de parcurgere a unui graf. Un algoritm de
    parcurgere este o funcție care primește un graf și un nod de start și
    întoarce o structură Traversal. Observați că tipul grafului este o
    variabilă de tip - algoritmii trebuie să funcționeze pentru orice
    structură de tip graf.
-}
type TraverseAlgo g = g -> Node -> Traversal

{-
    5. (2.5p) Implementați algoritmul de parcurgere în adâncime. (Depth-First Search)
-}
dfs :: Graph g => TraverseAlgo g
dfs g start = undefined -- TODO

{-
    6. (2.5p) Implementați algoritmul de parcurgere în lățime. (Breadth-First Search)
-}
bfs :: Graph g => TraverseAlgo g
bfs g start = undefined -- TODO

-------------------------------------------------------------------------------

{-
    7. (BONUS 0.5p) Definiți funcția sortDegree care primește ca parametru un graf
    și care întoarce un graf care are nodurile sortate descrescător în funcție de
    gradul exterior (numărul arcelor care ies din nodul respectiv).
    Observație: Dacă două noduri u și v (u < v) au același grad exterior, ele o să
    apară în ordinea u, v.
    Hint: Puteți folosi funcția sortBy
    Tipul lui sortBy: (a -> a -> Ordering) -> [a] -> [a]
    Exemplu de utilizare: sortBy compare [3,2,5,2,1] 
-}
sortDegree :: Graph g => g -> g
sortDegree = undefined -- TODO

-------------------------------------------------------------------------------

{-
    8. (BONUS 2.5p) Definiți tipul ColGraph care modelează un graf cu nodurile 
    colorate, fiind reprezentat ca o listă de tupluri de forma (nod, culoare) și 
    lista muchiilor. Includeți ColGraph în clasa Graph.

    Pentru a colora un graf, vom folosi algoritmul Welsh-Powell:
    * Pasul 1: Sortăm nodurile descrescător în funcție de valoarea gradului exterior.
    * Pasul 2: Extragem primul nod necolorat și îi atribuim prima culoare disponibila din
    lista possibleColors.
    * Pasul 3: Parcurgem lista cu nodurile rămase necolorate și verificăm care din ele
    se pot avea culoarea selectată la pasul anterior. Pentru ca două noduri să poată 
    avea aceeași culoare, trebuie ca ele să nu fie adiacente.
    * Pasul 4: Pașii 2 și 3 sunt reluați iterativ până ajungem să colorăm toate nodurile. 
-}

-- reprezentăm o culoare ca un String
type Color = String

-- lista de culori posibile
possibleColors :: [Color]
possibleColors = ["Red", "Green", "Blue", "Yellow", "Cyan", "Magenta", "Rose", "Violet", "Azure", "Orange"]

-- TODO: de rafinat constructorul de date AGraph
data ColGraph = CGraph Node deriving (Show)

instance Graph ColGraph where
    -- TODO

{-
    Definiți funcția care primește un graf, un nod și returnează culoarea nodului
    respectiv.
-}
getColor :: ColGraph -> Node -> Color
getColor = undefined -- TODO

-------------------------------------------------------------------------------

{-
    9. (BONUS 1p). Creați tipul GenericAdjListGraph care este este similar cu
    AdjListGraph, dar în care nodurile nu sunt neapărat de tip Int - pot
    fi de orice tip.
-}
-- TODO: de rafinat parametrul lui GALGraph
newtype GenericAdjListGraph a = GALGraph a deriving Show



{-
    10. (BONUS 1p). Adăugați tipul GenericAdjListGraph la clasa Functor
-}
instance Functor GenericAdjListGraph where
  -- TODO: de definit fmap
  fmap = undefined
